const { EventSource } = require('eventsource');

const Agent = require('../models/Agent');
const config = require('../config/keys');
const dashboardController = require('../controllers/dashboard');
const dockerController = require('../controllers/docker');
const { getCSTargetURL } = require('../utils/utility');

class JobEventsListener {
  constructor(sseEndpoint, options = {}) {
    this.sseEndpoint = sseEndpoint;
    this.eventSource = null;
    this.reconnectTimeout = options.reconnectTimeout || 10000; // Default reconnection time 10 seconds
    this.handlers = {
      // Handlers for incremental job updates
      'inc-job-created': options.onJobCreated || this.defaultHandler,
      'inc-job-updated': options.onJobUpdated || this.defaultHandler,
      'inc-job-deleted': options.onJobDeleted || this.defaultHandler,
      // Handlers for federated job updated
      'fl-job-created': options.onFLJobCreated || this.defaultHandler,
      'fl-job-updated': options.onFLJobUpdated || this.defaultHandler,
      'fl-job-deleted': options.onFLJobDeleted || this.defaultHandler,
    };
    this.isConnected = false;
    this.connectionListeners = [];
  }

  defaultHandler(data) {
    console.warn('Recevied event with no handler:', data);
  }

  connect() {
    this.log(
      `Listening for job updates stream in Central Service at ${this.sseEndpoint}...`
    );

    // Close existing connection if any
    if (this.eventSource) {
      this.eventSource.close();
    }

    this.eventSource = new EventSource(this.sseEndpoint);

    // Debug the event source state
    // console.log('EventSource readyState:', this.eventSource.readyState);

    // Handle connection 'open'
    this.eventSource.onopen = () => {
      this.log('Connection established with Central Service');
      this.isConnected = true;
    };

    // Generic message handler for other events
    this.eventSource.onmessage = (event) => {
      try {
        const data = JSON.parse(event.data);
        this.log('Received job updates from Central Service');

        if (data.type && this.handlers[data.type]) {
          this.handlers[data.type](data.job);
        }
      } catch (error) {
        this.error('Error processing event:', error);
      }
    };

    // Error handler
    this.eventSource.onerror = (error) => {
      this.error(`Error connecting to Central Service for job updates:`, error);

      // Close the current connection
      this.eventSource.close();
      this.isConnected = false;

      // Attempt to reconnect after a delay
      this.log(
        `Attempting to reconnect in ${this.reconnectTimeout / 1000} seconds...`
      );
      setTimeout(() => this.connect(), this.reconnectTimeout);
    };

    this.eventSource.addEventListener('heartbeat', (event) => {
      this.log('Received server heartbeat:', event.data);
    });

    return this;
  }

  disconnect() {
    if (this.eventSource) {
      console.log('Disconnecting from job updates stream');
      this.eventSource.close();
      this.eventSource = null;
      this.isConnected = false;
    }

    return this;
  }

  log(message, data = undefined) {
    if (data) console.log(`[SSE]: ${message}`, data);
    else console.log(`[SSE]: ${message}`);
  }

  error(message, error) {
    console.error(`[SSE]: ${message}`, error);
  }
}

const SSE_ENDPOINT = `${getCSTargetURL()}/api/events/${config.stationId}`

const jobListener = new JobEventsListener(SSE_ENDPOINT, {
  onFLJobCreated: async (job) => {
    console.log('Received FL Job request with payload', job);

    try {
      const agent = await Agent.findOne({
        type: 'federated',
        active: true,
        trainIds: { $in: [job.trainclassidlearning] },
      });

      // Skip job execution if no active agent
      if (!agent) {
        console.log(
          `[AGENT]: No active agent found for train "${job.trainclassidlearning}". Skipping job execution...`
        );
        return;
      }

      console.log(
        `[AGENT]: Found agent "${agent.name}" for train "${job.trainclassidlearning}". Starting job execution...`
      );

      // Pull train image
      await dockerController.pullFederatedImage(job.learningstoragelocation);

      console.log(
        `[AGENT]: Image pulled for train "${job.trainclassidlearning}". Started learning...`
      );

      const startLearningPayload = {
        pid: job.pid,
        jobid: job.jobid,
        learningstoragelocation: job.learningstoragelocation,
        trainclassidlearning: job.trainclassidlearning,
        currentround: parseInt(job.currentround),
        maxrounds: parseInt(job.maxrounds),
        bindMount: {
          hostSource: agent.hostSource,
          containerDestination: agent.containerDestination,
        },
        envs: agent.environmentVariables.map(({ name, type, value }) => ({
          name,
          type,
          value,
        })),
        privacysetting: agent.privacySetting,
        shmSize: '', // TODO: From Agent
      };

      // Start learning
      await dashboardController.startFederatedJobLearning(startLearningPayload);
    } catch (error) {
      console.error('Error executing federated learning job', error);
    }
  },
});

module.exports = jobListener;
