const request = require('request');

const config = require('../config/keys');
const FLJobManager = require('../federated/flJobManager');
const { FLJob, FLJobState, FLPrivacySetting } = require('../models/FLJob');
const { getAgentOptions } = require('../dind-certs-client');
const { vault, environment } = require('../utils');

async function readLabelForImage(image, callback) {
  //Use the filter parameter for the /image/ endpoint in the docker API
  //to only return the image we are interested in
  //See: https://docs.docker.com/engine/api/v1.43/#tag/Image
  //Filters are json objects
  const filter = {
    reference: [image],
  };

  //Build the request options and uri.
  //The filter object is encoded into a JSON string that will then be URL encoded
  let options = {
    uri: `https://${process.env.DOCKER_HOST}:${
      process.env.DOCKER_PORT
    }/images/json?all=1&filters=${encodeURIComponent(JSON.stringify(filter))}`,
    method: 'GET',
    agentOptions: getAgentOptions(),
  };

  request(options, function (error, response, body) {
    if (error) {
      console.log(error);
      callback(undefined);
    } else {
      let data = JSON.parse(body);
      if (data.length == 0) {
        console.log(`Found no image with name ${image}`);
        callback(undefined);
      } else {
        callback(JSON.stringify(data[0].Labels));
      }
    }
  });

  // TODO: Replace this above code with DockerClient here....
}

function parseEnvVarList(labels) {
  let envVariableList = [];

  if (labels) {
    let parsedLabels = JSON.parse(labels);
    if (parsedLabels && parsedLabels.envs)
      envVariableList = JSON.parse(parsedLabels.envs);
  }

  // Validate envs structure
  const input_type_list = ['number', 'password', 'text', 'url', 'select'];
  let index = envVariableList.length;
  let containerMount;
  while (index--) {
    const element = envVariableList[index];
    if (!element.name) {
      envVariableList.splice(index, 1);
      continue;
    }

    if (element.name === 'BIND_MOUNT_ENV') {
      containerMount = element.value;
      envVariableList.splice(index, 1);
      continue;
    }

    if (!input_type_list.includes(element.type)) {
      envVariableList[index].type = 'text';
    }

    if (!element.required) {
      envVariableList[index].required = false;
    }
  }

  return { envVariableList, containerMount };
}

function resolveVaultPathList(callback) {
  // key-value version 1 - START
  let kv1Promise = new Promise((resolve) => {
    //Get key-value engines (v1)
    vault.getKeyValueEngines(1).then((result) => {
      let promises = [];
      let kvEnginesVersion1 = result;
      kvEnginesVersion1.forEach((kvEngine) => {
        promises.push(
          new Promise((resolve) => {
            //Traverse all existing paths
            vault.traverse([kvEngine], (paths) => {
              resolve(paths);
            });
          })
        );
      });

      //When all results received
      Promise.all(promises).then((results) => {
        let paths = results.flat();
        //Read keys corresponding to each path
        let pathPromises = paths.map((path) => vault.read(path));
        Promise.all(pathPromises).then((results) => {
          let pathList = [];

          for (let index = 0; index < results.length; index++) {
            if (!results[index]) continue;
            const element = results[index];
            //key-value pairs
            const data = element.data;
            pathList.push(
              Object.keys(data).map((key) => {
                return { path: paths[index], key: key };
              })
            );
          }

          pathList = pathList.flat();
          resolve(pathList);
        });
      });
    });
  });
  // key-value version 1 - END

  // key-value version 2 - START
  let kv2Promise = new Promise((resolve) => {
    //Get key-value engines (v2)
    vault.getKeyValueEngines(2).then((result) => {
      let promises = [];
      let kvEnginesVersion2 = result;
      kvEnginesVersion2.forEach((kvEngine) => {
        promises.push(
          new Promise((resolve) => {
            //Traverse all existing paths (add 'metadata' for api calls)
            vault.traverse([`${kvEngine}metadata/`], (paths) => {
              resolve(paths);
            });
          })
        );
      });

      //When all results received
      Promise.all(promises).then((results) => {
        let paths = results.flat();

        //Replace 'metadata' with 'data' for api calls
        paths = paths.map((path) => {
          path = path.split('/');
          path.splice(1, 1, 'data');
          path = path.join('/');
          return path;
        });
        // console.log(paths);
        //Read keys corresponding to each path
        let pathPromises = paths.map((path) => vault.read(path));
        Promise.all(pathPromises).then((results) => {
          let pathList = [];

          for (let index = 0; index < results.length; index++) {
            if (!results[index]) continue;
            const element = results[index];
            //key-value pairs
            const data = element.data.data;
            pathList.push(
              Object.keys(data).map((key) => {
                return { path: paths[index], key: `data.${key}` };
              })
            );
          }

          pathList = pathList.flat();
          resolve(pathList);
        });
      });
    });
  });
  // key-value version 2 - END

  Promise.all([kv1Promise, kv2Promise]).then((results) => {
    callback(results.flat());
  });
}

module.exports = {
  // Create federated job container
  createFederatedJobContainer: async (payload) => {
    const {
      jobid,
      trainclassidlearning,
      learningstoragelocation,
      currentround,
      maxrounds,
      pid,
    } = payload;

    // Check vault status
    try {
      await vault.healthStatus();
    } catch (error) {
      throw new Error(`Error with vault. Check if it is unsealed: ${error}`);
    }

    // Create federated job container
    try {
      readLabelForImage(learningstoragelocation, (labels) => {
        if (!labels) {
          throw new Error(
            `Failed to load env variables from federated learning image: ${learningstoragelocation}`
          );
        }

        let { envVariableList, containerMount } = parseEnvVarList(labels);
        resolveVaultPathList((pathList) => {
          return {
            jobid,
            trainclassidlearning,
            learningstoragelocation,
            currentround,
            maxrounds,
            pid,
            envVariableList,
            paths: pathList,
            hostMount: config.hostMountDir || '',
            containerMount: containerMount || '',
          };
        });
      });
    } catch (error) {
      throw new Error(`Error creating federated job container: ${error}`);
    }
  },

  startFederatedJobLearning: async (payload) => {
    const {
      jobid,
      trainclassidlearning,
      learningstoragelocation,
      currentround,
      maxrounds,
      bindMount,
      shmSize,
      pid,
      envs,
      privacysetting,
    } = payload;

    // Check vault status
    try {
      await vault.healthStatus();
    } catch (error) {
      throw new Error(`Error with vault. Check if it is unsealed: ${error}`);
    }

    const job = await FLJob.findOne({ jobid: jobid });
    if (job !== null && job.state !== FLJobState.WAIT_FOR_ACCEPT) {
      //Job was already started
      console.log(`The job ${jobid} has already started learning.`);
      return;
    }

    //Load environment variables
    let envVariables;
    try {
      console.log('Creating environment variables...');
      envVariables = await environment.createEnvsArray(envs);
    } catch (error) {
      throw new Error(
        `Something went wrong while creating environment variables: ${error}`
      );
    }

    let mount = undefined;
    if (bindMount.hostSource && bindMount.containerDestination) {
      //Creates a binding string in the format expected by docker
      mount = `${bindMount.hostSource}:${bindMount.containerDestination}${
        bindMount.readOnly ? ':ro' : ''
      }`;
    }

    let sharedMemSize = 67108864; // 64 MB by default
    if (shmSize && shmSize > 0) {
      sharedMemSize = parseInt(shmSize) * 1048576; // shmSize is in MB, multiply to convert it to Bytes
    }

    //Create new worker
    const manager = new FLJobManager();
    const worker = await manager.createNewWorker(
      jobid,
      pid,
      trainclassidlearning,
      learningstoragelocation,
      currentround,
      maxrounds,
      envVariables,
      mount,
      sharedMemSize,
      privacysetting === 'inspect-results'
        ? FLPrivacySetting.INSPECT_RESULTS
        : FLPrivacySetting.DO_NOT_INSPECT_RESULTS
    );

    //Start the worker
    worker
      .start()
      .then(() => {
        console.log(`The job ${jobid} has started learning.`);
      })
      .catch((error) => {
        throw new Error(`The job ${jobid} could not start learning: ${error}`);
      });
  },
};
