const express = require('express');
const Joi = require('joi');

const Agent = require('../models/Agent');
const router = express.Router();

const agentSchema = Joi.object({
  name: Joi.string().required(),
  active: Joi.boolean().default(true),
  type: Joi.string().valid('incremental', 'federated').required(),
  privacySetting: Joi.string()
    .valid('inspect-results', 'no-result-inspection')
    .required(),
  trainIds: Joi.array().items(Joi.string()).required(),
  environmentVariables: Joi.array().items(
    Joi.object({
      name: Joi.string().required(),
      type: Joi.string().required(),
      value: Joi.string().required(),
    })
  ),
  executionParameters: Joi.object().pattern(Joi.string(), Joi.string()),
  inspectionBypass: Joi.boolean().default(false),
  hostSource: Joi.string().required(),
  containerDestination: Joi.string().required(),
});

/**
 * TODO: Add Keycloak middleware to all APIs
 */

module.exports = (keycloak) => {
  // Get all agents
  router.get('/', keycloak.protect(), async (req, res) => {
    try {
      const agents = await Agent.find();
      res.json(agents);
    } catch (err) {
      res.status(500).json({ message: err.message });
    }
  });

  // Get agent by ID
  router.get('/:id', [keycloak.protect(), getAgent], (req, res) => {
    res.json(res.agent);
  });

  // Create agent
  router.post('/', keycloak.protect(), async (req, res) => {
    const { error } = agentSchema.validate(req.body);
    if (error)
      return res
        .status(400)
        .json({ message: error.details[0].message, upper: true });

    const agent = new Agent({
      name: req.body.name,
      active: req.body.active,
      type: req.body.type,
      trainIds: req.body.trainIds,
      environmentVariables: req.body.environmentVariables,
      executionParameters: req.body.executionParameters,
      inspectionBypass: req.body.inspectionBypass,
      privacySetting: req.body.privacySetting,
      hostSource: req.body.hostSource,
      containerDestination: req.body.containerDestination,
    });

    try {
      const newAgent = await agent.save();
      res.status(201).json(newAgent);
    } catch (err) {
      let message = err.message;
      if (err?.errors?.name) {
        message = `${message}. ${err.errors.name}`;
      }
      res.status(400).json({ message, lower: true });
    }
  });

  // Update agent
  router.patch('/:id', [keycloak.protect(), getAgent], async (req, res) => {
    const { error } = agentSchema.validate(req.body);
    if (error)
      return res.status(400).json({ message: error.details[0].message });

    const updateableFields = [
      'name',
      'active',
      'type',
      'trainIds',
      'environmentVariables',
      'executionParameters',
      'inspectionBypass',
      'privacySetting',
      'hostSource',
      'containerDestination',
    ];

    updateableFields.forEach((field) => {
      if (req.body[field] != null) {
        res.agent[field] = req.body[field];
      }
    });

    try {
      const updatedAgent = await res.agent.save();
      res.json(updatedAgent);
    } catch (err) {
      res.status(400).json({ message: err.message });
    }
  });

  // Delete agent
  router.delete('/:id', [keycloak.protect(), getAgent], async (req, res) => {
    try {
      await res.agent.remove();
      res.json({ message: 'Deleted Agent' });
    } catch (err) {
      res.status(500).json({ message: err.message });
    }
  });

  // Middleware to get agent by ID
  async function getAgent(req, res, next) {
    const agentId = req.params.id;

    let agent;
    try {
      agent = await Agent.findById(agentId);
      if (agent == null) {
        return res
          .status(404)
          .json({ message: `Cannot find agent with id ${agentId}` });
      }
    } catch (err) {
      const message = `Error fetching agent with id: ${agentId}`;
      console.error(message, err);
      return res.status(500).json({ message });
    }

    res.agent = agent;
    next();
  }

  return router;
};
