import { Formik, Form } from 'formik';
import { Fragment, useState, useRef } from 'react';
import { useDispatch, useSelector } from 'react-redux';
import { grey } from '@mui/material/colors';
import List from '@mui/material/List';
import ListItem from '@mui/material/ListItem';
import ListItemText from '@mui/material/ListItemText';
import CircularProgress from '@mui/material/CircularProgress';
import Dialog from '@mui/material/Dialog';
import DialogContent from '@mui/material/DialogContent';
import DialogActions from '@mui/material/DialogActions';
import Grid from '@mui/material/Grid';
import IconButton from '@mui/material/IconButton';
import Typography from '@mui/material/Typography';
import TextField from '@mui/material/TextField';
import MenuItem from '@mui/material/MenuItem'
import FormControlLabel from '@mui/material/FormControlLabel';
import Switch from '@mui/material/Switch';
import Stack from '@mui/material/Stack';
import InfoIcon from '@mui/icons-material/InfoOutlined';
import AddIcon from '@mui/icons-material/AddCircleOutlineRounded';
import DeleteIcon from '@mui/icons-material/Delete';

import StyledButton from '../Button';
import StyledDialogTitle from '../DialogTitle';
import PrivacySetting from '../PrivacySetting';
import TextFieldWrapper from '../Form/TextField';
import AddCustomEnvForm from '../../core-ui/AddCustomEnvForm';
import { privacySettingOptions } from '../../constants';
import { getValidationProps } from '../../utils';
import { HtmlTooltip } from '../Tooltip';
import { fetchAgents, createAgent } from '../../redux/reducers/agentSlice';

export default function CreateAgentModal({ open, handleClose }) {
  /* Selectos */
  const dispatch = useDispatch();
  const { creating } = useSelector((state) => state.agent);

  /* State and Refs */
  const agentNameRef = useRef(null);
  const hostSourceRef = useRef(null);
  const containerDestRef = useRef(null);
  const [openAddEnv, setOpenAddEnv] = useState(false);
  const [newTrainId, setNewTrainId] = useState('');
  const [agentConfig, setAgentPayload] = useState({
    name: '',
    type: 'federated',
    active: false,
    hostSource: '',
    containerDestination: '',
    trainIds: [],
    environmentVariables: [],
    privacySetting: privacySettingOptions.INSPECT_RESULTS,
  });

  /* Constants */
  const trainIds = agentConfig.trainIds;
  const environmentVars = agentConfig.environmentVariables;
  const { initialFormState, FormSchema } = getValidationProps(environmentVars);

  const isInspectResults =
    agentConfig.privacySetting === privacySettingOptions.INSPECT_RESULTS;

  /* Handlers */
  const handleUpdateAgentConfig = ({ key, value }) => {
    setAgentPayload({ ...agentConfig, [key]: value });
  };

  const handleAddEnvVariables = (payload) => {
    agentConfig.environmentVariables.push(payload);
  };

  const handleAddTrainId = () => {
    if (!newTrainId) return;
    if (!trainIds.includes(newTrainId)) {
      const updatedTrainIds = [newTrainId, ...trainIds];
      handleUpdateAgentConfig({ key: 'trainIds', value: updatedTrainIds });

      setNewTrainId('');
    } else {
      alert('Train ID is already present');
    }
  };

  const handleRemoveTrainId = (index) => {
    const updatedTrainIds = trainIds.filter((_, i) => i !== index);
    handleUpdateAgentConfig({ key: 'trainIds', value: updatedTrainIds });
  };

  const handleCreateAgent = (formdata) => {
    const payload = {
      name: agentNameRef.current.value,
      active: agentConfig.active,
      hostSource: hostSourceRef.current.value,
      containerDestination: containerDestRef.current.value,
      type: agentConfig.type,
      privacySetting: agentConfig.privacySetting,
      trainIds: agentConfig.trainIds,
      environmentVariables: agentConfig.environmentVariables.map((env) => ({
        name: env.name,
        type: env.type,
        value: formdata[env.name],
      })),
    };

    dispatch(createAgent(payload)).then(() => {
      dispatch(fetchAgents());
      handleClose();
    });
  };

  const formId = 'create-agent-form';
  return (
    <Dialog fullWidth maxWidth="md" open={open}>
      <StyledDialogTitle onClose={handleClose}>
        Create New Agent
      </StyledDialogTitle>
      <DialogContent dividers>
        <Stack sx={{ mb: 2 }} gap={2}>
          {/* Basic Configuration */}
          <Typography variant="h6" sx={{ fontWeight: 'bold' }}>
            General
          </Typography>
          <Grid container spacing={2}>
            <Grid item xs={4}>
              <TextField
                fullWidth
                size="small"
                label="Agent Name"
                disabled={creating}
                inputRef={agentNameRef}
              />
            </Grid>
            <Grid item xs={4}>
              <TextField
                select
                fullWidth
                size="small"
                label="Agent Type"
                value={agentConfig.type}
                disabled={creating}
                onChange={({ target }) =>
                  handleUpdateAgentConfig({
                    key: 'type',
                    value: target.value,
                  })
                }
              >
                <MenuItem value="incremental">Incremental</MenuItem>
                <MenuItem value="federated">Federated</MenuItem>
              </TextField>
            </Grid>
            <Grid item xs={4}>
              <FormControlLabel
                control={
                  <Switch
                    checked={agentConfig.active}
                    color="success"
                    onChange={({ target }) =>
                      handleUpdateAgentConfig({
                        key: 'active',
                        value: target.checked,
                      })
                    }
                  />
                }
                label={<b>{agentConfig.active ? 'Active' : 'Inactive'}</b>}
              />
            </Grid>
          </Grid>
        </Stack>

        {/* Privacy Setting */}
        <PrivacySetting
          privacySetting={agentConfig.privacySetting}
          inspectResults={isInspectResults}
          disabled={creating}
          onChangePrivacy={({ target }) =>
            handleUpdateAgentConfig({
              key: 'privacySetting',
              value: target.value,
            })
          }
        />

        {/* Environment Variables */}
        <Stack
          direction="row"
          alignItems="center"
          justifyContent="space-between"
          my={2}
        >
          <Typography variant="h6" fontWeight="bold">
            Environment Variables
          </Typography>
          <HtmlTooltip placement="left" title="Add environment variable">
            <span>
              <IconButton
                color="inherit"
                onClick={() => setOpenAddEnv(!openAddEnv)}
              >
                <AddIcon />
              </IconButton>
            </span>
          </HtmlTooltip>
        </Stack>

        {/* Add custom ENVs form */}
        <AddCustomEnvForm
          open={openAddEnv}
          handleClose={() => setOpenAddEnv(false)}
          action={handleAddEnvVariables}
          formTitle="Add Environment Variable"
          skipExtra
        />

        {/* Check if ENVs present */}
        {environmentVars.length === 0 && (
          <Stack
            mb={4}
            direction="row"
            justifyContent="center"
            sx={{ p: 2, bgcolor: grey[100], borderRadius: 2 }}
          >
            <Typography>
              There are no environment variables for this agent. You can add
              using
            </Typography>
            <AddIcon sx={{ mx: 1 }} />
            <Typography>button above.</Typography>
          </Stack>
        )}

        {/* Custom ENV Form */}
        <Formik
          initialValues={initialFormState}
          validationSchema={FormSchema}
          onSubmit={handleCreateAgent}
        >
          <Form id={formId} noValidate>
            <Grid container spacing={2}>
              {environmentVars?.map((env, idx) => (
                <Fragment key={`agent-env-variable-${env?.name}-${idx}`}>
                  <Grid item xs={6}>
                    <TextField
                      fullWidth
                      size="small"
                      label="Name"
                      name="name"
                      value={env?.name}
                      disabled={Boolean(env?.disabled)}
                    />
                  </Grid>
                  <Grid item xs={6}>
                    <TextFieldWrapper
                      label="Value"
                      id={env?.name}
                      name={env?.name}
                      type={env?.type !== 'number' ? env?.type : undefined}
                      required={Boolean(env?.required)}
                      disabled={creating}
                    />
                  </Grid>
                </Fragment>
              ))}
            </Grid>
          </Form>
        </Formik>

        {/* Add Train IDs */}
        <Stack direction="row" spacing={1} sx={{ my: 2, alignItems: 'center' }}>
          <Typography variant="h6" sx={{ fontWeight: 'bold' }}>
            Train IDs
          </Typography>
        </Stack>
        <Stack direction="row" spacing={1} sx={{ my: 2, alignItems: 'center' }}>
          <TextField
            fullWidth
            size="small"
            variant="outlined"
            placeholder="Add trains starting with e.g.'federated_train_class_repository' or 'train_class_repository'"
            value={newTrainId}
            onChange={(e) => setNewTrainId(e.target.value)}
          />
          <HtmlTooltip placement="left" title="Add Train ID">
            <IconButton color="primary" onClick={handleAddTrainId}>
              <AddIcon />
            </IconButton>
          </HtmlTooltip>
        </Stack>

        {/* List Train IDs */}
        <List sx={{ maxHeight: 150, overflow: 'auto' }}>
          {agentConfig.trainIds.map((trainId, index) => (
            <ListItem
              key={trainId}
              secondaryAction={
                <IconButton
                  edge="end"
                  color="error"
                  aria-label="delete"
                  onClick={() => handleRemoveTrainId(index)}
                >
                  <DeleteIcon />
                </IconButton>
              }
            >
              <ListItemText primary={trainId} />
            </ListItem>
          ))}
        </List>

        {/* Bind Mounts */}
        <Stack direction="row" spacing={1} sx={{ my: 2, alignItems: 'center' }}>
          <Typography variant="h6" sx={{ fontWeight: 'bold' }}>
            Bind Mounts
          </Typography>
          <HtmlTooltip placement="right" title="Paths are case-sensitive">
            <InfoIcon />
          </HtmlTooltip>
        </Stack>

        {/* Bind Mounts Values */}
        <Grid container spacing={2}>
          <Grid item xs={6}>
            <HtmlTooltip
              title="Absolute path on Docker-in-Docker container from where the
                      file/directory needs to be mounted. The file/directory must
                      exist."
            >
              <TextField
                fullWidth
                inputRef={hostSourceRef}
                size="small"
                variant="outlined"
                label="Host Source Absolute Path"
                name="hostMount"
                color="success"
                defaultValue={agentConfig.hostSource}
                disabled={creating}
              />
            </HtmlTooltip>
          </Grid>
          <Grid item xs={6}>
            <HtmlTooltip title="Absolute path on the train container where the file/directory should be mounted.">
              <TextField
                fullWidth
                inputRef={containerDestRef}
                size="small"
                variant="outlined"
                label="Container Destination Absolute Path"
                name="containerMount"
                color="success"
                defaultValue={agentConfig.containerDestination}
                disabled={creating}
              />
            </HtmlTooltip>
          </Grid>
        </Grid>
      </DialogContent>
      <DialogActions>
        <StyledButton variant="contained" type="submit" form={formId}>
          {creating ? <CircularProgress size={25} thickness={5} /> : 'Create'}
        </StyledButton>
      </DialogActions>
    </Dialog>
  );
}
