import { Formik, Form } from 'formik';
import { Fragment, useState, useRef } from 'react';
import { useDispatch, useSelector } from 'react-redux';
import { grey } from '@mui/material/colors';
import List from '@mui/material/List';
import ListItem from '@mui/material/ListItem';
import ListItemText from '@mui/material/ListItemText';
import CircularProgress from '@mui/material/CircularProgress';
import Dialog from '@mui/material/Dialog';
import DialogContent from '@mui/material/DialogContent';
import DialogActions from '@mui/material/DialogActions';
import Grid from '@mui/material/Grid';
import IconButton from '@mui/material/IconButton';
import Typography from '@mui/material/Typography';
import TextField from '@mui/material/TextField';
import MenuItem from '@mui/material/MenuItem';
import FormControlLabel from '@mui/material/FormControlLabel';
import Switch from '@mui/material/Switch';
import Stack from '@mui/material/Stack';
import InfoIcon from '@mui/icons-material/InfoOutlined';
import AddIcon from '@mui/icons-material/AddCircleOutlineRounded';
import DeleteIcon from '@mui/icons-material/Delete';

import StyledButton from '../Button';
import StyledDialogTitle from '../DialogTitle';
import TextFieldWrapper from '../Form/TextField';
import PrivacySetting from '../PrivacySetting';
import AddCustomEnvForm from '../../core-ui/AddCustomEnvForm';
import { privacySettingOptions } from '../../constants';
import { HtmlTooltip } from '../Tooltip';
import { getValidationProps } from '../../utils';
import {
  fetchAgents,
  updateAgent,
  updateAgentConfig,
  addCustomEnvVariable,
} from '../../redux/reducers/agentSlice';

export default function EditAgentModal({ open, handleClose }) {
  /* Selectos */
  const dispatch = useDispatch();
  const { selectedAgent, updating } = useSelector((state) => state.agent);

  /* State and Refs */
  const agentNameRef = useRef(null);
  const hostSourceRef = useRef(null);
  const containerDestRef = useRef(null);
  const [newTrainId, setNewTrainId] = useState('');
  const [openAddEnv, setOpenAddEnv] = useState(false);

  /* Constants */
  const environmentVars = selectedAgent?.environmentVariables;
  const trainIds = selectedAgent?.trainIds;
  const isInspectResults =
    selectedAgent?.privacySetting === privacySettingOptions.INSPECT_RESULTS;
  const { initialFormState, FormSchema } = getValidationProps(
    environmentVars || []
  );

  /* Handlers */
  const handleAddTrainId = () => {
    if (!newTrainId) return;
    if (!trainIds.includes(newTrainId)) {
      const updatedTrainIds = [newTrainId, ...trainIds];
      dispatch(updateAgentConfig({ key: 'trainIds', value: updatedTrainIds }));
      setNewTrainId('');
    } else {
      alert('Train ID is already present');
    }
  };

  const handleRemoveTrainId = (index) => {
    const updatedTrainIds = trainIds.filter((_, i) => i !== index);
    dispatch(updateAgentConfig({ key: 'trainIds', value: updatedTrainIds }));
  };

  const handleUpdateAgentConfig = ({ key, value }) => {
    dispatch(updateAgentConfig({ key, value }));
  };

  const handleUpdateAgent = (formdata) => {
    const payload = {
      agentId: selectedAgent._id,
      payload: {
        name: agentNameRef.current.value,
        active: selectedAgent.active,
        hostSource: hostSourceRef.current.value,
        containerDestination: containerDestRef.current.value,
        type: selectedAgent.type,
        privacySetting: selectedAgent.privacySetting,
        trainIds: selectedAgent.trainIds,
        environmentVariables: selectedAgent.environmentVariables.map((env) => ({
          name: env.name,
          type: env.type,
          value: formdata[env.name],
        })),
      },
    };

    dispatch(updateAgent(payload)).then(() => {
      dispatch(fetchAgents());
      handleClose();
    });
  };

  const formId = 'edit-agent-form';
  return (
    <Dialog fullWidth maxWidth="md" open={open}>
      <StyledDialogTitle onClose={handleClose}>
        Edit Agent Configuration
      </StyledDialogTitle>
      <DialogContent dividers>
        <Stack spacing={1} sx={{ p: 2, bgcolor: grey[100], borderRadius: 2 }}>
          <Typography>
            <b>Agent ID: </b>
            {selectedAgent?._id}
          </Typography>
        </Stack>
        {/* Basic Configuration */}
        <Stack sx={{ my: 2 }} gap={2}>
          <Typography variant="h6" sx={{ fontWeight: 'bold' }}>
            General
          </Typography>
          <Grid container spacing={2}>
            <Grid item xs={4}>
              <TextField
                fullWidth
                size="small"
                label="Agent Name"
                disabled={updating}
                inputRef={agentNameRef}
                defaultValue={selectedAgent?.name}
              />
            </Grid>
            <Grid item xs={4}>
              <TextField
                select
                fullWidth
                size="small"
                label="Agent Type"
                value={selectedAgent?.type}
                disabled={updating}
                onChange={({ target }) =>
                  handleUpdateAgentConfig({
                    key: 'type',
                    value: target.value,
                  })
                }
              >
                <MenuItem value="incremental">Incremental</MenuItem>
                <MenuItem value="federated">Federated</MenuItem>
              </TextField>
            </Grid>
            <Grid item xs={4}>
              <FormControlLabel
                control={
                  <Switch
                    checked={selectedAgent?.active}
                    color="success"
                    onChange={({ target }) =>
                      handleUpdateAgentConfig({
                        key: 'active',
                        value: target.checked,
                      })
                    }
                  />
                }
                label={<b>{selectedAgent?.active ? 'Active' : 'Inactive'}</b>}
              />
            </Grid>
          </Grid>
        </Stack>

        {/* Privacy Setting */}
        <PrivacySetting
          privacySetting={selectedAgent?.privacySetting}
          inspectResults={isInspectResults}
          disabled={updating}
          onChangePrivacy={({ target }) =>
            handleUpdateAgentConfig({
              key: 'privacySetting',
              value: target.value,
            })
          }
        />

        {/* Environment Variables */}
        <Stack
          direction="row"
          alignItems="center"
          justifyContent="space-between"
          my={2}
        >
          <Typography variant="h6" fontWeight="bold">
            Environment Variables
          </Typography>
          <HtmlTooltip placement="left" title="Add environment variable">
            <span>
              <IconButton
                color="inherit"
                onClick={() => setOpenAddEnv(!openAddEnv)}
              >
                <AddIcon />
              </IconButton>
            </span>
          </HtmlTooltip>
        </Stack>

        {/* Add custom ENVs form */}
        <AddCustomEnvForm
          open={openAddEnv}
          handleClose={() => setOpenAddEnv(false)}
          action={addCustomEnvVariable}
          formTitle="Add Environment Variable"
          skipExtra
        />

        {/* Check if ENVs present */}
        {environmentVars?.length === 0 && (
          <Stack
            mb={4}
            direction="row"
            justifyContent="center"
            sx={{ p: 2, bgcolor: grey[100], borderRadius: 2 }}
          >
            <Typography>
              There are no environment variables for this agent. You can add
              using
            </Typography>
            <AddIcon sx={{ mx: 1 }} />
            <Typography>button above.</Typography>
          </Stack>
        )}

        {/* Custom ENV Form */}
        <Formik
          initialValues={initialFormState}
          validationSchema={FormSchema}
          onSubmit={handleUpdateAgent}
        >
          {({ setFieldValue }) => (
            <Form id={formId} noValidate>
              <Grid container spacing={2}>
                {environmentVars?.map((env, idx) => (
                  <Fragment key={`agent-env-variable-${env?.name}-${idx}`}>
                    <Grid item xs={6}>
                      <TextField
                        fullWidth
                        size="small"
                        label="Name"
                        name="name"
                        value={env?.name}
                        disabled={Boolean(env?.disabled)}
                      />
                    </Grid>
                    <Grid item xs={6}>
                      <TextFieldWrapper
                        label="Value"
                        id={env?.name}
                        name={env?.name}
                        type={env?.type !== 'number' ? env?.type : undefined}
                        required={Boolean(env?.required)}
                        disabled={updating}
                      />
                    </Grid>
                  </Fragment>
                ))}
              </Grid>
            </Form>
          )}
        </Formik>

        {/* Add Train IDs */}
        <Stack direction="row" spacing={1} sx={{ my: 2, alignItems: 'center' }}>
          <Typography variant="h6" sx={{ fontWeight: 'bold' }}>
            Train IDs
          </Typography>
        </Stack>
        <Stack direction="row" spacing={1} sx={{ my: 2, alignItems: 'center' }}>
          <TextField
            fullWidth
            size="small"
            variant="outlined"
            placeholder="Add trains starting with e.g.'federated_train_class_repository' or 'train_class_repository'"
            value={newTrainId}
            onChange={(e) => setNewTrainId(e.target.value)}
          />
          <HtmlTooltip placement="left" title="Add Train ID">
            <IconButton color="primary" onClick={handleAddTrainId}>
              <AddIcon />
            </IconButton>
          </HtmlTooltip>
        </Stack>

        {/* List Train IDs */}
        <List sx={{ maxHeight: 150, overflow: 'auto' }}>
          {selectedAgent?.trainIds?.map((trainId, index) => (
            <ListItem
              key={trainId}
              secondaryAction={
                <IconButton
                  edge="end"
                  color="error"
                  aria-label="delete"
                  onClick={() => handleRemoveTrainId(index)}
                >
                  <DeleteIcon />
                </IconButton>
              }
            >
              <ListItemText primary={trainId} />
            </ListItem>
          ))}
        </List>

        {/* Bind Mounts */}
        <Stack direction="row" spacing={1} sx={{ my: 2, alignItems: 'center' }}>
          <Typography variant="h6" sx={{ fontWeight: 'bold' }}>
            Bind Mounts
          </Typography>
          <HtmlTooltip placement="right" title="Paths are case-sensitive">
            <InfoIcon />
          </HtmlTooltip>
        </Stack>

        {/* Bind Mounts Values */}
        <Grid container spacing={2}>
          <Grid item xs={6}>
            <HtmlTooltip
              title="Absolute path on Docker-in-Docker container from where the
                      file/directory needs to be mounted. The file/directory must
                      exist."
            >
              <TextField
                fullWidth
                inputRef={hostSourceRef}
                size="small"
                variant="outlined"
                label="Host Source Absolute Path"
                name="hostMount"
                color="success"
                defaultValue={selectedAgent?.hostSource}
                disabled={updating}
              />
            </HtmlTooltip>
          </Grid>
          <Grid item xs={6}>
            <HtmlTooltip title="Absolute path on the train container where the file/directory should be mounted.">
              <TextField
                fullWidth
                inputRef={containerDestRef}
                size="small"
                variant="outlined"
                label="Container Destination Absolute Path"
                name="containerMount"
                color="success"
                defaultValue={selectedAgent?.containerDestination}
                disabled={updating}
              />
            </HtmlTooltip>
          </Grid>
        </Grid>

        {/* TODO: Shared Memory */}
        <Typography variant="h6" sx={{ fontWeight: 'bold', my: 2 }}>
          Shared Memory Size
        </Typography>
        <Grid container spacing={2}>
          <Grid item xs={6}>
            <HtmlTooltip
              title="Size of /dev/shm in MB. The size must be greater than 0. If
                      omitted the system uses 64MB."
            >
              <TextField
                fullWidth
                size="small"
                variant="outlined"
                label="Size of /dev/shm in MB"
                type="number"
                color="success"
                disabled={updating}
                inputProps={{ min: 0 }}
                // value={shmSize}
                // onChange={handleChangeShmSize}
              />
            </HtmlTooltip>
          </Grid>
        </Grid>
      </DialogContent>
      <DialogActions>
        <StyledButton variant="contained" type="submit" form={formId}>
          {updating ? <CircularProgress size={25} thickness={5} /> : 'Update'}
        </StyledButton>
      </DialogActions>
    </Dialog>
  );
}
