import { useState, useEffect } from 'react';
import { useDispatch, useSelector } from 'react-redux';
import Button from '@mui/material/Button';
import IconButton from '@mui/material/IconButton';
import EditIcon from '@mui/icons-material/Edit';
import Stack from '@mui/material/Stack';
import TableRow from '@mui/material/TableRow';
import TableCell from '@mui/material/TableCell';
import Typography from '@mui/material/Typography';
import DeleteIcon from '@mui/icons-material/Delete';
import CheckIcon from '@mui/icons-material/CheckCircle';
import CancelIcon from '@mui/icons-material/Cancel';
import AddIcon from '@mui/icons-material/AddCircleOutlineRounded';

import AgentTable from '../../components/Table';
import EditAgentModal from '../../components/edit-agent-modal';
import CreateAgentModal from '../../components/create-agent-model';
import { Icon007 } from './icon';
import { columns } from '../../constants/agent';
import { HtmlTooltip } from '../../components/Tooltip';
import {
  fetchAgents,
  deleteAgent,
  setSelectedAgent,
} from '../../redux/reducers/agentSlice';

const BYPASS_INSPECTION = 'no-result-inspection';

export default function Agent() {
  const dispatch = useDispatch();
  const [openEditAgent, setOpenEditAgent] = useState(false);
  const [openCreateAgent, setOpenCreateAgent] = useState(false);
  const { agents, loading } = useSelector((state) => state.agent);

  useEffect(() => {
    dispatch(fetchAgents());
  }, [dispatch]);

  const handleEditAgent = (row) => (_) => {
    dispatch(setSelectedAgent(row));
    setOpenEditAgent(true);
  };

  const handleDeleteImage = (row) => (_) => {
    dispatch(deleteAgent(row._id)).then(() => {
      dispatch(fetchAgents());
    });
  };

  const renderTableRows = (row, idx) => (
    <TableRow hover key={row?._id}>
      <TableCell>{idx + 1}</TableCell>
      <TableCell>{row?.name}</TableCell>
      <TableCell>
        <b>
          {row?.trainIds.slice(0, 2).join('\n')} <br />
          {row?.trainIds.length > 2 && '...'}
        </b>
      </TableCell>
      <TableCell>
        {row?.active ? (
          <CheckIcon sx={{ color: 'success.light' }} />
        ) : (
          <CancelIcon sx={{ color: 'error.light' }} />
        )}
      </TableCell>
      <TableCell>{row?.type}</TableCell>
      <TableCell>
        {row?.privacySetting === BYPASS_INSPECTION ? 'yes' : 'no'}
      </TableCell>
      <TableCell>{new Date(row?.createdAt).toLocaleString('de')}</TableCell>
      <TableCell>{new Date(row?.updatedAt).toLocaleString('de')}</TableCell>
      <TableCell>
        <HtmlTooltip title="Edit Agent">
          <span>
            <IconButton
              color="info"
              disabled={loading}
              onClick={handleEditAgent(row)}
            >
              <EditIcon />
            </IconButton>
          </span>
        </HtmlTooltip>
        <HtmlTooltip title="Delete Agent">
          <span>
            <IconButton
              color="error"
              disabled={loading}
              onClick={handleDeleteImage(row)}
            >
              <DeleteIcon />
            </IconButton>
          </span>
        </HtmlTooltip>
      </TableCell>
    </TableRow>
  );

  return (
    <>
      <Stack
        direction="row"
        my={3}
        alignItems="center"
        justifyContent="space-between"
      >
        <Stack direction="row" gap={1}>
          <Typography variant="h4" fontWeight="bold">
            Agent <em>007</em>
          </Typography>
          <Icon007 />
        </Stack>
        <Button
          variant="outlined"
          startIcon={<AddIcon />}
          onClick={() => setOpenCreateAgent(true)}
        >
          Create New Agent
        </Button>
      </Stack>
      <AgentTable
        rows={agents}
        columns={columns}
        renderTableRows={renderTableRows}
        loading={loading}
        emptyRowLabel={
          <>
            No data available in table for <b>agents</b>
          </>
        }
      />
      <CreateAgentModal
        open={openCreateAgent}
        handleClose={() => setOpenCreateAgent(false)}
      />
      <EditAgentModal
        open={openEditAgent}
        handleClose={() => setOpenEditAgent(false)}
      />
    </>
  );
}
