import { createSlice, createAsyncThunk } from '@reduxjs/toolkit';

import { showAlert } from './alertSlice';
import * as API from '../api/agent';

const initialState = {
  loading: false,
  updating: false,
  creating: false,
  error: '',
  agents: [],
  selectedAgent: {},
};

const agentSlice = createSlice({
  name: 'agent',
  initialState,
  reducers: {
    setSelectedAgent: (state, action) => {
      state.selectedAgent = action.payload;
    },
    updateAgentConfig: (state, { payload }) => {
      state.selectedAgent[payload.key] = payload.value;
    },
    addCustomEnvVariable: (state, { payload }) => {
      state.selectedAgent.environmentVariables.push(payload);
    },
  },
  extraReducers: (builder) => {
    builder
      .addCase(fetchAgents.pending, (state) => {
        state.loading = true;
      })
      .addCase(fetchAgents.fulfilled, (state, action) => {
        state.loading = false;
        state.agents = action.payload;
      })
      .addCase(fetchAgents.rejected, (state, action) => {
        state.loading = false;
        state.error = action.payload;
      })
      .addCase(deleteAgent.pending, (state) => {
        state.loading = true;
      })
      .addCase(deleteAgent.fulfilled, (state) => {
        state.loading = false;
      })
      .addCase(deleteAgent.rejected, (state) => {
        state.loading = false;
      })
      .addCase(createAgent.pending, (state) => {
        state.creating = true;
      })
      .addCase(createAgent.fulfilled, (state) => {
        state.creating = false;
      })
      .addCase(createAgent.rejected, (state) => {
        state.creating = false;
      })
      .addCase(updateAgent.pending, (state) => {
        state.updating = true;
      })
      .addCase(updateAgent.fulfilled, (state) => {
        state.updating = false;
      })
      .addCase(updateAgent.rejected, (state) => {
        state.updating = false;
      });
  },
});

export const fetchAgents = createAsyncThunk(
  'agent/fetchAgents',
  async (_, { dispatch, rejectWithValue }) => {
    try {
      const response = await API.fetchAgents();
      return response.data;
    } catch (err) {
      const message = `Unable to fetch agents: ${err.message}`;
      dispatch(
        showAlert({
          message: err?.response?.data?.message || message,
          options: {
            key: 'agent/fetchAgentsError',
            variant: 'error',
          },
        })
      );
      console.error(message);
      return rejectWithValue(message);
    }
  }
);

export const createAgent = createAsyncThunk(
  'agent/createAgent',
  async (payload, { dispatch, rejectWithValue }) => {
    try {
      await API.createAgent(payload);
      dispatch(
        showAlert({
          message: 'Agent created successfully',
          options: {
            key: 'agent/createAgentSuccess',
            variant: 'success',
          },
        })
      );
    } catch (err) {
      const message = `Error creating agent: ${err.message}`;
      dispatch(
        showAlert({
          message: err?.response?.data?.message || message,
          options: {
            key: 'agent/createAgentError',
            variant: 'error',
          },
        })
      );
      console.error(message);
      return rejectWithValue(message);
    }
  }
);

export const updateAgent = createAsyncThunk(
  'agent/updateAgent',
  async (payload, { dispatch, rejectWithValue }) => {
    try {
      await API.updateAgent(payload);
      dispatch(
        showAlert({
          message: `Agent ${payload.agentId} updated successfully`,
          options: {
            key: 'agent/updateAgentSuccess',
            variant: 'success',
          },
        })
      );
    } catch (err) {
      const message = `Error updating agent: ${err.message}`;
      dispatch(
        showAlert({
          message: err?.response?.data?.message || message,
          options: {
            key: 'agent/updateAgentError',
            variant: 'error',
          },
        })
      );
      console.error(message);
      return rejectWithValue(message);
    }
  }
);

export const deleteAgent = createAsyncThunk(
  'agent/deleteAgent',
  async (agentId, { dispatch, rejectWithValue }) => {
    try {
      await API.deleteAgent(agentId);
      dispatch(
        showAlert({
          message: `Agent ${agentId} deleted successfully`,
          options: {
            key: 'agent/deleteAgentSuccess',
            variant: 'success',
          },
        })
      );
    } catch (err) {
      const message = `Error deleting agent: ${err.message}`;
      dispatch(
        showAlert({
          message: err?.response?.data?.message || message,
          options: {
            key: 'agent/deleteAgentError',
            variant: 'error',
          },
        })
      );
      console.error(message);
      return rejectWithValue(message);
    }
  }
);

export const { addCustomEnvVariable, updateAgentConfig, setSelectedAgent } =
  agentSlice.actions;

export default agentSlice.reducer;
